import logging

import yaml

from n2y.utils import pandoc_ast_to_markdown, fromisoformat, sanitize_filename
from n2y.property_values import TitlePropertyValue


logger = logging.getLogger(__name__)


class Page:
    def __init__(self, client, notion_data):
        logger.debug("Instantiating page")
        self.client = client

        self.notion_id = notion_data['id']
        self.created_time = fromisoformat(notion_data['created_time'])
        self.created_by = client.wrap_notion_user(notion_data['created_by'])
        self.last_edited_time = fromisoformat(notion_data['last_edited_time'])
        self.last_edited_by = client.wrap_notion_user(notion_data['last_edited_by'])
        self.archived = notion_data['archived']
        self.icon = notion_data['icon'] and client.wrap_notion_file(notion_data['icon'])
        self.cover = notion_data['cover'] and client.wrap_notion_file(notion_data['cover'])
        self.archived = notion_data['archived']
        self.properties = {
            k: client.wrap_notion_property_value(npv)
            for k, npv in notion_data['properties'].items()
        }
        self.notion_parent = notion_data['parent']
        self.notion_url = notion_data['url']

        self._block = None
        self._children = None

    @property
    def title(self):
        for property_value in self.properties.values():
            # Notion ensure's there is always exactly one title property
            if isinstance(property_value, TitlePropertyValue):
                return property_value.rich_text

    @property
    def filename(self):
        # TODO: switch to using the database's natural keys as the file names
        return sanitize_filename(self.title.to_plain_text())

    @property
    def block(self):
        if self._block is None:
            self._block = self.client.get_block(self.notion_id, page=self)
        return self._block

    @property
    def children(self):
        if self._children is None:
            self._children = True  # TODO: Implement this
            raise NotImplementedError()
            # recursively look through blocks for child_pages and child_databases,
            # creating them in order that they appear in the block heirarchy
        return self._children

    @property
    def parent(self):
        parent_type = self.notion_parent["type"]
        if parent_type == "workspace":
            return None
        elif parent_type == "page_id":
            return self.client.get_page(self.notion_parent["page_id"])
        else:
            assert parent_type == "database_id"
            return self.client.get_database(self.notion_parent["database_id"])

    def to_pandoc(self):
        return self.block.to_pandoc()

    def content_to_markdown(self):
        pandoc_ast = self.to_pandoc()
        if pandoc_ast is not None:
            return pandoc_ast_to_markdown(pandoc_ast)
        else:
            return None

    def properties_to_values(self):
        return {k: v.to_value() for k, v in self.properties.items()}

    def to_markdown(self):
        return '\n'.join([
            '---',
            yaml.dump(self.properties_to_values()) + '---',
            self.content_to_markdown() or '',
        ])



from itertools import groupby
import logging
from urllib.parse import urljoin

from pandoc.types import (
    Str, Para, Plain, Header, CodeBlock, BulletList, OrderedList, Decimal,
    Period, Meta, Pandoc, Link, HorizontalRule, BlockQuote, Image, MetaString,
    Table, TableHead, TableBody, TableFoot, RowHeadColumns, Row, Cell, RowSpan,
    ColSpan, ColWidthDefault, AlignDefault, Caption, Math, DisplayMath,
)


logger = logging.getLogger(__name__)


class Block:
    """
    A Notion page's content consists of a tree of nested Blocks.

    See here for a listing of all of them: https://developers.notion.com/reference/block

    This class that wraps the Notion blocks and also is responsible for
    transforming them into Pandoc abstract syntax tree. A single Notion block
    often has multiple lines of text.

    Pandoc's data model doesn't line up exactly with Notion's, for example
    Notion doesn't have a wrapper around lists, while Pandoc does.
    """

    def __init__(self, client, notion_data, page=None, get_children=True):
        """
        The Notion client object is passed down for the following reasons:
        1. Some child objects may be unknown until the block is processed.
           Links to other Notion pages are an example.
        2. In some cases a block may choose not to get child blocks.
           Currently, all blocks load all children.
        """
        logger.debug('Instantiating "%s" block', type(self).__name__)
        self.client = client
        self.page = page

        self.notion_id = notion_data['id']
        self.created_time = notion_data['created_time']
        self.created_by = notion_data['created_by']
        self.last_edited_time = notion_data['last_edited_time']
        self.last_edited_by = notion_data['last_edited_by']
        self.has_children = notion_data['has_children']
        self.archived = notion_data['archived']
        self.notion_type = notion_data['type']
        self.notion_data = notion_data[notion_data['type']]

        if get_children:
            if self.has_children:
                children = self.client.get_child_blocks(self.notion_id, page, get_children)
            else:
                children = []
        else:
            children = None
        self.children = children

    def to_pandoc(self):
        raise NotImplementedError()

    def children_to_pandoc(self):
        assert self.has_children
        pandoc_ast = []
        for block_type, blocks in groupby(self.children, lambda c: type(c)):
            if issubclass(block_type, ListItemBlock):
                pandoc_ast.append(block_type.list_to_pandoc(blocks))
            else:
                for b in blocks:
                    result = b.to_pandoc()
                    if isinstance(result, list):
                        # a few blocks return lists of nodes
                        pandoc_ast.extend(result)
                    elif result is not None:
                        # a plugin may decide to return None to indicate the
                        # block should be removed; ideally pandoc.types.Nil
                        # would handle this, but it doesn't appear to work
                        pandoc_ast.append(result)
        return pandoc_ast

    @property
    def notion_url(self):
        # the notion URL's don't work if the dashes from the block ID are present
        fragment = '#' + self.notion_id.replace('-', '')
        if self.page is None:
            return fragment
        else:
            return urljoin(self.page.notion_url, fragment)


class ListItemBlock(Block):
    @classmethod
    def list_to_pandoc(klass, items):
        raise NotImplementedError()


class ChildPageBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.title = self.notion_data["title"]

    def to_pandoc(self):
        assert self.children is not None
        if self.children:
            children = self.children_to_pandoc()
            return Pandoc(Meta({'title': MetaString(self.title)}), children)
        else:
            return None


class EquationBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.expression = self.notion_data["expression"]

    def to_pandoc(self):
        return Para([Math(DisplayMath(), self.expression)])


class ParagraphBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(self.notion_data["rich_text"])

    def to_pandoc(self):
        content = self.rich_text.to_pandoc()
        if self.has_children:
            # Notion allows you to create child blocks for a paragraph; these
            # child blocks appear indented relative to the paragraph. There's
            # no way to represent this indentation in pandoc's AST, so we just
            # append the child blocks afterwards.
            result = [Para(content)]
            children = self.children_to_pandoc()
            result.extend(children)
        else:
            result = Para(content)
        return result


class BulletedListItemBlock(ListItemBlock):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(self.notion_data["rich_text"])

    def to_pandoc(self):
        content = [Plain(self.rich_text.to_pandoc())]
        if self.has_children:
            children = self.children_to_pandoc()
            content.extend(children)
        return content

    @classmethod
    def list_to_pandoc(klass, items):
        return BulletList([b.to_pandoc() for b in items])


class ToDoListItemBlock(BulletedListItemBlock):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.checked = self.notion_data['checked']

        # TODO: Move this into the "to_pandoc" stage
        box = '☒ ' if self.checked else '☐ '
        self.rich_text.prepend(box)


class NumberedListItemBlock(ListItemBlock):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(self.notion_data["rich_text"])

    def to_pandoc(self):
        content = [Plain(self.rich_text.to_pandoc())]
        if self.has_children:
            children = self.children_to_pandoc()
            content.extend(children)
        return content

    @classmethod
    def list_to_pandoc(klass, items):
        return OrderedList((1, Decimal(), Period()), [b.to_pandoc() for b in items])


class HeadingBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(self.notion_data["rich_text"])

        # The Notion UI allows one to bold the text in a header, but the bold
        # styling isn't displayed. Thus, to avoid unexpected appearances of
        # bold text in the generated documents, bolding is removed.
        for rich_text in self.rich_text:
            rich_text.bold = False

    def to_pandoc(self):
        return Header(self.level, ('', [], []), self.rich_text.to_pandoc())


class HeadingOneBlock(HeadingBlock):
    level = 1


class HeadingTwoBlock(HeadingBlock):
    level = 2


class HeadingThreeBlock(HeadingBlock):
    level = 3


class DividerBlock(Block):
    def to_pandoc(self):
        return HorizontalRule()


class BookmarkBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.url = self.notion_data["url"]
        self.caption = client.wrap_notion_rich_text_array(self.notion_data["caption"])

    def to_pandoc(self):
        if self.caption:
            caption_ast = self.caption.to_pandoc()
        else:
            caption_ast = [Str(self.url)]
        return Para([Link(('', [], []), caption_ast, (self.url, ''))])


class FencedCodeBlock(Block):
    pandoc_highlight_languages = [
        "abc", "actionscript", "ada", "agda", "apache", "asn1", "asp", "ats", "awk",
        "bash", "bibtex", "boo", "c", "changelog", "clojure", "cmake", "coffee",
        "coldfusion", "comments", "commonlisp", "cpp", "cs", "css", "curry", "d",
        "default", "diff", "djangotemplate", "dockerfile", "dot", "doxygen",
        "doxygenlua", "dtd", "eiffel", "elixir", "elm", "email", "erlang", "fasm",
        "fortranfixed", "fortranfree", "fsharp", "gcc", "glsl", "gnuassembler", "go",
        "graphql", "groovy", "hamlet", "haskell", "haxe", "html", "idris", "ini",
        "isocpp", "j", "java", "javadoc", "javascript", "javascriptreact", "json",
        "jsp", "julia", "kotlin", "latex", "lex", "lilypond", "literatecurry",
        "literatehaskell", "llvm", "lua", "m4", "makefile", "mandoc", "markdown",
        "mathematica", "matlab", "maxima", "mediawiki", "metafont", "mips", "modelines",
        "modula2", "modula3", "monobasic", "mustache", "nasm", "nim", "noweb",
        "objectivec", "objectivecpp", "ocaml", "octave", "opencl", "orgmode", "pascal",
        "perl", "php", "pike", "postscript", "povray", "powershell", "prolog",
        "protobuf", "pure", "purebasic", "python", "qml", "r", "raku", "relaxng",
        "relaxngcompact", "rest", "rhtml", "roff", "ruby", "rust", "sass", "scala",
        "scheme", "sci", "scss", "sed", "sgml", "sml", "spdxcomments", "sql",
        "sqlmysql", "sqlpostgresql", "stan", "stata", "swift", "systemverilog", "tcl",
        "tcsh", "texinfo", "toml", "typescript", "verilog", "vhdl", "xml", "xorg",
        "xslt", "xul", "yacc", "yaml", "zsh",
    ]

    # TODO: finish filling in this mapping from Notion language names to
    # pandoc's supported language names
    notion_to_pandoc_highlight_languages = {
        'c#': 'cs',
        'c++': 'cpp',
        'f#': 'fsharp',
        'objective-c': 'objectivec',
        'docker': 'dockerfile',
        'coffee': 'coffeescript',
        'shell': 'bash',  # seems better than nothing
    }

    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.language = self.notion_data["language"]
        self.rich_text = client.wrap_notion_rich_text_array(self.notion_data["rich_text"])
        self.caption = client.wrap_notion_rich_text_array(self.notion_data["caption"])

    def to_pandoc(self):
        pandoc_language = self.notion_to_pandoc_highlight_languages.get(
            self.language, self.language,
        )
        if pandoc_language not in self.pandoc_highlight_languages:
            if pandoc_language != "plain text":
                msg = 'Dropping syntax highlighting for unsupported language "%s" (%s)'
                logger.warning(msg, pandoc_language, self.notion_url)
            language = []
        else:
            language = [pandoc_language]
        return CodeBlock(('', language, []), self.rich_text.to_plain_text())


class QuoteBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(self.notion_data["rich_text"])

    def to_pandoc(self):
        return BlockQuote([Para(self.rich_text.to_pandoc())])


class FileBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.file = client.wrap_notion_file(notion_data['file'])
        self.caption = client.wrap_notion_rich_text_array(self.notion_data["caption"])

    def to_pandoc(self):
        url = None
        if self.file.type == "external":
            url = self.file.url
        elif self.file.type == "file":
            url = self.client.download_file(self.file.url, self.page)
        if self.caption:
            caption_ast = self.caption.to_pandoc()
        else:
            caption_ast = [Str(url)]
        return Para([Link(('', [], []), caption_ast, (url, ''))])


class ImageBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.file = client.wrap_notion_file(notion_data['image'])
        self.caption = client.wrap_notion_rich_text_array(self.notion_data["caption"])

    def to_pandoc(self):
        url = None
        if self.file.type == "external":
            url = self.file.url
        elif self.file.type == "file":
            url = self.client.download_file(self.file.url, self.page)
        return Para([Image(('', [], []), self.caption.to_pandoc(), (url, ''))])


class TableBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.has_column_header = self.notion_data['has_column_header']
        self.has_row_header = self.notion_data['has_row_header']
        self.table_width = self.notion_data['table_width']

    def to_pandoc(self):
        children = self.children_to_pandoc()
        # Isolate the header row if it exists
        if self.has_column_header:
            header_rows = [children.pop(0)]
        else:
            header_rows = []
        if self.has_row_header:
            row_header_columns = 1
        else:
            row_header_columns = 0
        # Notion does not have cell alignment or width options, sticking with defaults.
        colspec = [(AlignDefault(), ColWidthDefault()) for _ in range(self.table_width)]
        table = Table(
            ('', [], []),
            Caption(None, []),
            colspec,
            TableHead(('', [], []), header_rows),
            [TableBody(
                ('', [], []),
                RowHeadColumns(row_header_columns),
                [],
                children
            )],
            TableFoot(('', [], []), [])
        )
        return table


class RowBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.cells = [client.wrap_notion_rich_text_array(nc) for nc in self.notion_data["cells"]]

    def to_pandoc(self):
        cells = [Cell(
            ('', [], []),
            AlignDefault(),
            RowSpan(1),
            ColSpan(1),
            [Plain(cell.to_pandoc())]
        ) for cell in self.cells]
        return Row(('', [], []), cells)


class ToggleBlock(Block):
    """
    Generates a bulleted list item with indented children. A plugin may be used
    to add html classes and replicate the interactive behavior found in Notion.
    """

    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(self.notion_data["rich_text"])

    def to_pandoc(self):
        header = self.rich_text.to_pandoc()
        children = self.children_to_pandoc()
        content = [Para(header)]
        content.extend(children)
        return BulletList([content])


class CalloutBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        super().__init__(client, notion_data, page, get_children)
        self.rich_text = client.wrap_notion_rich_text_array(self.notion_data["rich_text"])
        # the color and icon are not currently used

    def to_pandoc(self):
        content = self.rich_text.to_pandoc()
        if self.has_children:
            children = self.children_to_pandoc()
            result = [Para(content)]
            result.extend(children)
        else:
            result = Para(content)
        return result


class NoopBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        # don't get the child blocks, as we're not using the data
        super().__init__(client, notion_data, page, get_children=False)

    def to_pandoc(self):
        return None


class TableOfContentsBlock(NoopBlock):
    pass


class BreadcrumbBlock(NoopBlock):
    pass


class UnsupportedBlock(NoopBlock):
    pass


class TemplateBlock(NoopBlock):
    pass


class WarningBlock(NoopBlock):
    def to_pandoc(self):
        logger.warning('Skipping unsupported "%s" block (%s)', self.notion_type, self.notion_url)
        return None


class ChildDatabaseBlock(WarningBlock):
    pass


class EmbedBlock(WarningBlock):
    pass


class VideoBlock(WarningBlock):
    pass


class PdfBlock(WarningBlock):
    pass


class ColumnBlock(WarningBlock):
    pass


class ColumnListBlock(WarningBlock):
    pass


class LinkPreviewBlock(WarningBlock):
    pass


class SyncedBlock(Block):
    def __init__(self, client, notion_data, page, get_children=True):
        self.original = notion_data[notion_data["type"]]["synced_from"] is None
        super().__init__(client, notion_data, page, get_children=self.original)
        # Synced blocks will always have children unless not shared
        # (There will always be at least one UnsupportedBlock child)
        self.shared = self.has_children
        self.children = self._get_synced_block_children()

    def _get_synced_block_children(self):
        if not self.original and self.shared:
            return self.client.get_child_blocks(
                self.notion_data["synced_from"]["block_id"],
                self.page, True,
            )
        return self.children

    def to_pandoc(self):
        if not self.shared:
            logger.warning('Skipping un-shared synced block (%s)', self.notion_url)
            return None
        return self.children_to_pandoc()


class LinkToPageBlock(WarningBlock):
    pass


DEFAULT_BLOCKS = {
    "paragraph": ParagraphBlock,
    "heading_1": HeadingOneBlock,
    "heading_2": HeadingTwoBlock,
    "heading_3": HeadingThreeBlock,
    "bulleted_list_item": BulletedListItemBlock,
    "numbered_list_item": NumberedListItemBlock,
    "to_do": ToDoListItemBlock,
    "toggle": ToggleBlock,
    "child_page": ChildPageBlock,
    "child_database": ChildDatabaseBlock,
    "embed": EmbedBlock,
    "image": ImageBlock,
    "video": VideoBlock,
    "file": FileBlock,
    "pdf": PdfBlock,
    "bookmark": BookmarkBlock,
    "callout": CalloutBlock,
    "quote": QuoteBlock,
    "equation": EquationBlock,
    "divider": DividerBlock,
    "table_of_contents": TableOfContentsBlock,
    "breadcrumb": TableOfContentsBlock,
    "column": ColumnBlock,
    "column_list": ColumnListBlock,
    "link_preview": LinkPreviewBlock,
    "synced_block": SyncedBlock,
    "template": TemplateBlock,
    "link_to_page": LinkToPageBlock,
    "code": FencedCodeBlock,
    "table": TableBlock,
    "table_row": RowBlock,
    "unsupported": UnsupportedBlock,
}



import logging


logger = logging.getLogger(__name__)


class File:
    """
    See https://developers.notion.com/reference/file-object
    """

    def __init__(self, client, notion_data):
        self.client = client
        if notion_data['type'] == "file":
            logger.debug('Instantiating file "%s"', notion_data['file']['url'])
            self.type = "file"
            self.url = notion_data['file']['url']
            self.expiry_time = notion_data['file']['expiry_time']
        elif notion_data['type'] == "external":
            logger.debug('Instantiating external file "%s"', notion_data['external']['url'])
            self.type = "external"
            self.url = notion_data['external']['url']
        else:
            file_type = notion_data['type']
            raise ValueError(f'Unknown file type "{file_type}"')



class Property:
    def __init__(self, client, notion_data):
        self.client = client
        self.notion_id = notion_data['id']
        self.notion_type = notion_data['type']
        self.name = notion_data['name']


# TODO: eventually add methods for generating SQL type expresions or JSON
# schema from these guys


class TitleProperty(Property):
    pass


class TextProperty(Property):
    pass


class NumberProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.format = notion_data['number']["format"]


class SelectProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        notion_options = notion_data['select']['options']
        self.options = [SelectOption(self.client, no) for no in notion_options]


class SelectOption:
    def __init__(self, client, notion_option):
        self.client = client
        self.notion_id = notion_option['id']
        self.name = notion_option['name']
        self.color = notion_option['color']


class MultiSelectProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        notion_options = notion_data['multi_select']['options']
        self.options = [MultiSelectOption(self.client, no) for no in notion_options]


class MultiSelectOption:
    """
    This class appears identical to SelectOption, but the notion API docs calls it a
    separate thing so it appears like it may change in the future; also, it
    seems like one may want a plugin that treats the two types of options
    differently
    """

    def __init__(self, client, notion_option):
        self.client = client
        self.notion_id = notion_option['id']
        self.name = notion_option['name']
        self.color = notion_option['color']


class DateProperty(Property):
    pass


class PeopleProperty(Property):
    pass


class FilesProperty(Property):
    pass


class CheckboxProperty(Property):
    pass


class UrlProperty(Property):
    pass


class EmailProperty(Property):
    pass


class PhoneNumberProperty(Property):
    pass


class FormulaProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.formula = notion_data["formula"]["expression"]


class RelationProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        relation = notion_data["relation"]
        self.database_id = relation["database_id"]
        self.synced_property_name = relation.get("synced_property_name", None)
        self.synced_property_id = relation.get("synced_property_id", None)


class RollupProperty(Property):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        rollup = notion_data["rollup"]
        self.relation_property_name = rollup["relation_property_name"]
        self.relation_property_id = rollup["relation_property_id"]
        self.rollup_property_name = rollup["rollup_property_name"]
        self.rollup_property_id = rollup["rollup_property_id"]
        self.function = rollup["function"]


class CreatedTimeProperty(Property):
    pass


class CreatedByProperty(Property):
    pass


class LastEditedTimeProperty(Property):
    pass


class LastEditedBy(Property):
    pass


DEFAULT_PROPERTIES = {
    'title': TitleProperty,
    'rich_text': TextProperty,
    'number': NumberProperty,
    'select': SelectProperty,
    'multi_select': MultiSelectProperty,
    'date': DateProperty,
    'people': PeopleProperty,
    'files': FilesProperty,
    'checkbox': CheckboxProperty,
    'url': UrlProperty,
    'email': EmailProperty,
    'phone_number': PhoneNumberProperty,
    'formula': FormulaProperty,
    'relation': RelationProperty,
    'rollup': RollupProperty,
    'created_time': CreatedTimeProperty,
    'created_by': CreatedByProperty,
    'last_edited_time': LastEditedTimeProperty,
    'last_edited_by': LastEditedBy,
}



from datetime import datetime
import logging
import re

import pandoc
from pandoc.types import Str, Space
from plumbum import ProcessExecutionError

from n2y.errors import PandocASTParseError


logger = logging.getLogger(__name__)


def process_notion_date(notion_date):
    if notion_date is None:
        return None
    elif notion_date.get('end', None):
        return [
            notion_date['start'],
            notion_date['end'],
        ]
    else:
        return notion_date['start']


def processed_date_to_plain_text(processed_date):
    # TODO: make this work the same way that Notion does
    if processed_date is None:
        return ""
    if isinstance(processed_date, list):
        return f'{processed_date[0]} to {processed_date[1]}'
    else:
        return processed_date


# see https://pandoc.org/MANUAL.html#exit-codes
PANDOC_PARSE_ERROR = 64


def pandoc_ast_to_markdown(pandoc_ast):
    # This function tries to avoid calling the separate pandoc binary (which is
    # slow) for basic cases with just spaces and strings
    if pandoc_ast is None or pandoc_ast == []:
        return ""
    elif type(pandoc_ast) == list and all(type(n) in [Str, Space] for n in pandoc_ast):
        # TODO: optimize performance for some other basic cases
        return ''.join(
            ' ' if isinstance(n, Space) else n[0]
            for n in pandoc_ast
        )
    else:
        try:
            # TODO: add a mechanism to customize this
            result = pandoc.write(
                pandoc_ast,
                format='gfm+tex_math_dollars+raw_attribute',
                options=[
                    '--wrap', 'none',  # don't hard line-wrap
                    '--eol', 'lf',  # use linux-style line endings
                ],
            )
        except ProcessExecutionError as err:
            if err.retcode == PANDOC_PARSE_ERROR:
                lines = []
                # TODO: update this code to make it not print so much
                for element, path in pandoc.iter(pandoc_ast, path=True):
                    path_str = ".".join(str(i) for _, i in path)
                    lines.append(f"{path_str} {element}")
                    logger.error("Pandoc AST:\n%s", "\n".join(lines))
                msg = (
                    "Pandoc couldn't parse the generated AST. "
                    f"This is likely due to a bug in n2y or a plugin: {err.stderr}"
                )
                logger.error(msg)
                raise PandocASTParseError(msg)
        return result


def fromisoformat(datestring):
    """
    Parse Notion's datestrings, which aren't handled out of the box by
    `datetime.fromisoformat` because of the `Z` at the end of them.

    This function removes the need for a third-party library.
    """
    if datestring.endswith('Z'):
        return datetime.fromisoformat(datestring[:-1] + '+00:00')
    else:
        return datetime.fromisoformat(datestring)


def sanitize_filename(filename):
    """Taken from django."""
    s = str(filename).strip().replace(" ", "_")
    s = re.sub(r"(?u)[^-\w.]", "", s)
    if s in {".", ".."}:
        raise ValueError("Could not derive file name from '%s'" % filename)
    return s


def strip_dashes(string):
    return string.replace("-", "")



import logging

import yaml

from n2y.property_values import RelationPropertyValue
from n2y.utils import fromisoformat, sanitize_filename


logger = logging.getLogger(__name__)


class Database:
    def __init__(self, client, notion_data):
        logger.debug("Instantiating database")
        self.client = client

        self.notion_id = notion_data['id']
        self.created_time = fromisoformat(notion_data['created_time'])
        self.created_by = client.wrap_notion_user(notion_data['created_by'])
        self.last_edited_time = fromisoformat(notion_data['last_edited_time'])
        self.last_edited_by = client.wrap_notion_user(notion_data['last_edited_by'])
        self.title = client.wrap_notion_rich_text_array(notion_data['title'])
        self.icon = notion_data['icon'] and client.wrap_notion_file(notion_data['icon'])
        self.cover = notion_data['cover'] and client.wrap_notion_file(notion_data['cover'])
        self.archived = notion_data['archived']
        self.schema = {
            k: client.wrap_notion_property(p)
            for k, p in notion_data['properties'].items()
        }
        self.notion_parent = notion_data['parent']
        self.notion_url = notion_data['url']
        self.archived = notion_data['archived']

        self._children = None

    @property
    def filename(self):
        return sanitize_filename(self.title.to_plain_text())

    @property
    def children(self):
        if self._children is None:
            self._children = self.client.get_database_pages(self.notion_id)
        return self._children

    @property
    def parent(self):
        if self.notion_parent["type"] == "workspace":
            return None
        else:
            return self.client.get_page(self.notion_parent["page_id"])

    @property
    def related_database_ids(self):
        """
        This method is much more complicated than it should be due to
        limitations of the Notion API.

        First, one would expect that the RelationProperty objects would be
        present in the databases's properties features, however they do not
        show up _unless_ the relationship is back to the same database.

        Secondly, one would expect that the page property endpoint
        (https://developers.notion.com/reference/retrieve-a-page-property)
        would enable one to retrieve the related database id from the property
        directly, however, the database id doesn't appear to be returned there
        either.

        As a last result, this method will first get the first page in a
        database (raising an error if there are no pages). Then, it will loop
        through the properties of the page to find any relationship properties.
        Then, it will loop through all pages in the database to find one that
        actually has a value of a page in the related database. Finally, we
        retrieve the related page and get the database ID from the parent.
        """
        database_title = self.title.to_plain_text()
        if len(self.children) == 0:
            logger.error(
                'Unable to identify relationships for empty database "%s"',
                database_title
            )
        first_page = self.children[0]
        ids = []
        for prop_name, prop in first_page.properties.items():
            if isinstance(prop, RelationPropertyValue):
                related_page_id = None
                for page in self.children:
                    related_page_ids = page.properties[prop_name].ids
                    if len(related_page_ids) > 0:
                        related_page_id = related_page_ids[0]
                        break
                if related_page_id is None:
                    logger.error(
                        'Unable to identify related database for relationship "%s" '
                        'property in the "%s" database because there are no values '
                        'in the entire database (%s)',
                        prop_name,
                        database_title,
                        self.notion_url,
                    )
                else:
                    related_page = self.client.get_page(related_page_id)
                    assert related_page.notion_parent["type"] == "database_id"
                    ids.append(related_page.notion_parent["database_id"])
        return ids

    def to_pandoc(self):
        return self.block.to_pandoc()

    def to_yaml(self):
        content_property = self.client.content_property
        id_property = self.client.id_property
        if content_property in self.schema:
            logger.warning(
                'The content property "%s" is shadowing an existing '
                'property with the same name', content_property,
            )
        if id_property in self.schema:
            logger.warning(
                'The id property "%s" is shadowing an existing '
                'property with the same name', content_property,
            )
        results = []
        for page in self.children:
            result = page.properties_to_values()
            if content_property:
                content = page.content_to_markdown()
                result[content_property] = content
            if id_property:
                notion_id = page.notion_id
                result[id_property] = notion_id
            results.append(result)
        return yaml.dump(results, sort_keys=False)



import hashlib
import logging
import json
from os import path, makedirs
import os
import shutil
import tempfile
from urllib.parse import urljoin, urlparse
import importlib.util

import requests

from n2y.errors import (
    HTTPResponseError, APIResponseError, ObjectNotFound, PluginError,
    UseNextClass, is_api_error_code, APIErrorCode
)
from n2y.file import File
from n2y.page import Page
from n2y.database import Database
from n2y.blocks import DEFAULT_BLOCKS
from n2y.properties import DEFAULT_PROPERTIES
from n2y.property_values import DEFAULT_PROPERTY_VALUES
from n2y.user import User
from n2y.rich_text import DEFAULT_RICH_TEXTS, RichTextArray
from n2y.mentions import DEFAULT_MENTIONS
from n2y.utils import strip_dashes


DEFAULT_NOTION_CLASSES = {
    "page": Page,
    "database": Database,
    "blocks": DEFAULT_BLOCKS,
    "properties": DEFAULT_PROPERTIES,
    "property_values": DEFAULT_PROPERTY_VALUES,
    "user": User,
    "file": File,
    "rich_text_array": RichTextArray,
    "rich_texts": DEFAULT_RICH_TEXTS,
    "mentions": DEFAULT_MENTIONS,
}


logger = logging.getLogger(__name__)


class Client:
    """
    An instance of the client class has a few purposes:

    1. To store configuration
    2. To retrieve data from Notion
    3. To determine what classes to use to wrap this notion data, based on the configuration
    4. To act as a shared global store for all of the objects that are pulled
       from Notion.

    In particular there is a cache of all pages and databases which ensure that
    the database class and page class are instantiated exactly once for each
    page or database.
    """

    def __init__(
        self,
        access_token,
        media_root='.',
        media_url='',
        plugins=None,
        content_property=None,
        id_property=None,
        database_config=None,
    ):
        self.access_token = access_token
        self.media_root = media_root
        self.media_url = media_url
        self.content_property = content_property
        self.id_property = id_property
        self.database_config = database_config if database_config is not None else {}

        self.base_url = "https://api.notion.com/v1/"
        self.headers = {
            "Authorization": f"Bearer {self.access_token}",
            "Content-Type": "application/json",
            "Notion-Version": "2022-02-22",
        }

        self.databases_cache = {}
        self.pages_cache = {}

        self.notion_classes = self.get_default_classes()
        self.load_plugins(plugins)
        self.plugin_data = {}

    def get_default_classes(self):
        notion_classes = {}
        for notion_object, object_types in DEFAULT_NOTION_CLASSES.items():
            if type(object_types) == dict:
                notion_classes[notion_object] = {k: [v] for k, v in object_types.items()}
            else:
                notion_classes[notion_object] = [object_types]
        return notion_classes

    def load_plugins(self, plugins):
        if plugins is not None:
            for plugin in plugins:
                plugin_module = importlib.import_module(plugin)
                try:
                    self.load_plugin(plugin_module.notion_classes)
                except PluginError as err:
                    logger.error('Error loading plugin "%s": %s', plugin, err)
                    raise

    def load_plugin(self, notion_classes):
        for notion_object, object_types in notion_classes.items():
            if notion_object in self.notion_classes:
                default_object_types = DEFAULT_NOTION_CLASSES[notion_object]
                self._override_notion_classes(notion_object, object_types, default_object_types)
            else:
                raise PluginError(f'Invalid notion object "{notion_object}"')

    def _override_notion_classes(self, notion_object, object_types, default_object_types):
        # E.g., there are many types of notion blocks but only one type of notion page.
        notion_object_has_types = isinstance(default_object_types, dict)

        if notion_object_has_types and isinstance(object_types, dict):
            for object_type, plugin_class in object_types.items():
                if object_type in default_object_types:
                    class_being_replaced = default_object_types[object_type]
                    # assumes all of the default classes have a single parent class
                    base_class = class_being_replaced.__bases__[0]
                    if issubclass(plugin_class, base_class):
                        self.notion_classes[notion_object][object_type].append(plugin_class)
                    else:
                        raise PluginError(
                            f'Cannot use "{plugin_class.__name__}", as it doesn\'t '
                            f'override the base class "{base_class.__name__}"',
                        )
                else:
                    raise PluginError(f'Invalid type "{object_type}" for "{notion_object}"')
        elif notion_object_has_types and not isinstance(object_types, dict):
            raise PluginError(f'Expecting dict for "{notion_object}", found "{type(object_types)}"')
        else:
            plugin_class = object_types
            base_class = default_object_types
            if issubclass(plugin_class, base_class):
                self.notion_classes[notion_object].append(plugin_class)
            else:
                raise PluginError(
                    f'Cannot use "{plugin_class.__name__}", as it doesn\'t '
                    f'override the base class "{base_class.__name__}"',
                )

    def get_class_list(self, notion_object, object_type=None):
        if object_type is None:
            return self.notion_classes[notion_object]
        try:
            return self.notion_classes[notion_object][object_type]
        except KeyError:
            raise NotImplementedError(f'Unknown "{notion_object}" class of type "{object_type}"')

    def instantiate_class(self, notion_object, object_type, *args, **kwargs):
        class_list = self.get_class_list(notion_object, object_type)
        for klass in reversed(class_list):
            try:
                return klass(*args, **kwargs)
            except UseNextClass:
                logger.debug("Skipping %s due to UseNextClass exception", klass.__name__)

    def _wrap_notion_page(self, notion_data):
        """
        Wrap notion page data in the appropriate class. If the page class has
        already been created, then ignore the new notion data and just return
        the existing version.

        Note that it may seem redundant to check the cache here, since it's also
        checked in `get_page`, however unlike databases which are only ever
        retrieved via `get_database`, pages can be retrieved either through
        `get_page` or through `get_database_pages`. Thus, it's possible that a
        database page be first retrieved individually and then retrieved via the
        full database. In this case, it's unfortunate (but unavoidable) that we
        retrieved the data from Notion twice, but even so we don't want to
        replace our existing page instance, along with it's content or other
        state that has been added to it.
        """
        if notion_data["id"] in self.pages_cache:
            return self.pages_cache[notion_data["id"]]
        else:
            page = self.instantiate_class("page", None, self, notion_data)
            self.pages_cache[page.notion_id] = page
            return page

    def _wrap_notion_database(self, notion_data):
        return self.instantiate_class("database", None, self, notion_data)

    def wrap_notion_block(self, notion_data, page, get_children):
        return self.instantiate_class(
            "blocks", notion_data["type"],
            self, notion_data, page, get_children,
        )

    def wrap_notion_user(self, notion_data):
        return self.instantiate_class("user", None, self, notion_data)

    def wrap_notion_file(self, notion_data):
        return self.instantiate_class("file", None, self, notion_data)

    def wrap_notion_rich_text_array(self, notion_data):
        return self.instantiate_class("rich_text_array", None, self, notion_data)

    def wrap_notion_rich_text(self, notion_data):
        return self.instantiate_class("rich_texts", notion_data["type"], self, notion_data)

    def wrap_notion_mention(self, notion_data, plain_text):
        # here we pass in the plain_text to avoid the need to query the page
        # just to get its title
        return self.instantiate_class(
            "mentions", notion_data["type"],
            self, notion_data, plain_text,
        )

    def wrap_notion_property(self, notion_data):
        return self.instantiate_class("properties", notion_data["type"], self, notion_data)

    def wrap_notion_property_value(self, notion_data):
        return self.instantiate_class("property_values", notion_data["type"], self, notion_data)

    def get_page_or_database(self, object_id):
        """
        First attempt to get the page corresponding with the object id; if
        the page doesn't exist, then attempt to retrieve the database. This
        trial-and-error is necessary because the API doesn't provide a means to
        determining what type of object corresponds with an ID and we don't want
        to make the user indicate if they are pulling down a database or a page.
        """
        page = self.get_page(object_id)
        if page is not None:
            return page
        else:
            return self.get_database(object_id)

    def get_database(self, database_id):
        """
        Retrieve the database (but not it's pages) if its not in the cache. Even
        if it is in the cache.
        """
        if database_id in self.databases_cache:
            database = self.databases_cache[database_id]
        else:
            try:
                notion_database = self._get_url(f"{self.base_url}databases/{database_id}")
                database = self._wrap_notion_database(notion_database)
            except ObjectNotFound:
                database = None
            self.databases_cache[database_id] = database
        return database

    def get_database_pages(self, database_id):
        notion_pages = self.get_database_notion_pages(database_id)
        return [self._wrap_notion_page(np) for np in notion_pages]

    def get_database_notion_pages(self, database_id):
        results = []
        url = f"{self.base_url}databases/{database_id}/query"
        request_data = self._create_database_request_data(database_id)
        while True:
            data = self._post_url(url, request_data)
            results.extend(data["results"])
            if not data["has_more"]:
                return results
            else:
                request_data["start_cursor"] = data["next_cursor"]

    def _create_database_request_data(self, dashed_database_id):
        database_id = strip_dashes(dashed_database_id)
        return self.database_config[database_id] if database_id in self.database_config else {}

    def get_page(self, page_id):
        """
        Retrieve the page if its not in the cache.
        """
        if page_id in self.pages_cache:
            page = self.pages_cache[page_id]
        else:
            try:
                notion_page = self._get_url(f"{self.base_url}pages/{page_id}")
            except ObjectNotFound:
                self.pages_cache[page_id] = None
                return
            # _wrap_notion_page will add the page to the cache
            page = self._wrap_notion_page(notion_page)
        return page

    def get_block(self, block_id, page, get_children=True):
        notion_block = self.get_notion_block(block_id)
        return self.wrap_notion_block(notion_block, page, get_children)

    def get_notion_block(self, block_id):
        url = f"{self.base_url}blocks/{block_id}"
        response = requests.get(url, headers=self.headers)
        return self._parse_response(response)

    def get_child_blocks(self, block_id, page, get_children):
        child_notion_blocks = self.get_child_notion_blocks(block_id)
        return [self.wrap_notion_block(b, page, get_children) for b in child_notion_blocks]

    def get_child_notion_blocks(self, block_id):
        url = f"{self.base_url}blocks/{block_id}/children"
        params = {}
        results = []
        while True:
            data = self._get_url(url, params)
            results.extend(data["results"])
            if not data["has_more"]:
                return results
            else:
                params["start_cursor"] = data["next_cursor"]

    def get_page_property(self, page_id, property_id):
        notion_property = self.get_notion_page_property(page_id, property_id)
        return self.wrap_notion_property(notion_property)

    def get_notion_page_property(self, page_id, property_id):
        url = f"{self.base_url}pages/{page_id}/properties/{property_id}"
        response = requests.get(url, headers=self.headers)
        return self._parse_response(response)

    def _get_url(self, url, params=None):
        if params is None:
            params = {}
        response = requests.get(url, headers=self.headers, params=params)
        return self._parse_response(response)

    def _post_url(self, url, data=None):
        if data is None:
            data = {}
        response = requests.post(url, headers=self.headers, json=data)
        return self._parse_response(response)

    def _parse_response(self, response):
        """Taken from https://github.com/ramnes/notion-sdk-py"""
        try:
            response.raise_for_status()
        except requests.HTTPError as error:
            try:
                body = error.response.json()
                code = body.get("code")
            except json.JSONDecodeError:
                code = None
            if code == APIErrorCode.ObjectNotFound:
                raise ObjectNotFound(response, body["message"])
            elif code and is_api_error_code(code):
                raise APIResponseError(response, body["message"], code)
            raise HTTPResponseError(error.response)
        return response.json()

    def download_file(self, url, page):
        """
        Download a file from a given URL into the MEDIA_ROOT.

        Preserve the file extension from the URL, but use the name of the parent
        page followed by an md5 hash.
        """
        url_path = path.basename(urlparse(url).path)
        _, extension = path.splitext(url_path)
        with requests.get(url, stream=True) as request_stream:
            content_iterator = request_stream.iter_content(4096)
            return self.save_file(content_iterator, page, extension)

    def save_file(self, content_iterator, page, extension):
        """
        Save the content in the provided iterator into a file in MEDIA_ROOT. The
        file name is determined from the page name, file extension, and an md5
        hash of the content. The md5 hash is calculated as the data is streamed
        to a temporary file, which is then moved to the final location once the
        md5 hash can be calculated.
        """
        temp_fd, temp_filepath = tempfile.mkstemp()
        hash_md5 = hashlib.md5()
        with os.fdopen(temp_fd, 'wb') as temp_file:
            for chunk in content_iterator:
                hash_md5.update(chunk)
                temp_file.write(chunk)

        num_hash_characters = 8  # just long enough to avoid collisions
        hash = hash_md5.hexdigest()[:num_hash_characters]
        relative_filepath = "".join([page.filename, "-", hash, extension])
        full_filepath = path.join(self.media_root, relative_filepath)

        makedirs(path.dirname(full_filepath), exist_ok=True)
        shutil.move(temp_filepath, full_filepath)
        return urljoin(self.media_url, relative_filepath)


def id_from_share_link(share_link):
    hyphens_removed = share_link.replace("-", "")
    if not hyphens_removed.startswith("https://www.notion.so/"):
        return hyphens_removed
    else:
        return hyphens_removed.split("/")[-1].split("?")[0]



"""
A setuptools based setup module.
"""

from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

description = 'Notion to YAML'

setup(
    name='n2y',
    version='0.4.2',
    description=description,
    long_description=description,
    long_description_content_type='text/x-rst',
    url='https://github.com/innolitics/n2y',
    author='Innolitics, LLC',
    author_email='info@innolitics.com',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    keywords='notion documentation yaml markdown',
    packages=find_packages(exclude=['tests']),
    install_requires=['pyyaml', 'requests', 'pandoc'],
    extras_require={
        'dev': [
            'pytest',
            'flake8',
            'check-manifest',
        ]
    },
    data_files=[],
    entry_points={
        'console_scripts': [
            'n2y = n2y.main:cli_main'
        ]
    },
)



import os
import sys
import json
import logging
import argparse

from n2y import notion
from n2y.database import Database
from n2y.page import Page
from n2y.errors import APIErrorCode, APIResponseError

logger = None


def cli_main():
    args = sys.argv[1:]
    access_token = os.environ.get('NOTION_ACCESS_TOKEN', None)
    sys.exit(main(args, access_token))


def main(raw_args, access_token):
    parser = argparse.ArgumentParser(
        description="Move data from Notion into YAML/markdown",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument("object_id", help="The id or url for a Notion database or page")
    parser.add_argument(
        "--format", '-f',
        choices=["yaml", "yaml-related", "markdown"], default="yaml",
        help=(
            "Select output type (only applies to databases)\n"
            "  yaml - log yaml to stdout\n"
            "  yaml-related - save all related databases to a set of YAML files\n"
            "  markdown - create a markdown file for each page"
        )
    )
    parser.add_argument(
        "--content-property", default='',
        help=(
            "Store each database page's content in this property. "
            "The page's content isn't exported if it's set to a blank string. "
            "Only applies when dumping a database to YAML."
        )
    )
    parser.add_argument(
        "--id-property", default='id',
        help=(
            "Store each database page's id in this property. "
            "The page's id isn't exported if it's set to a blank string. "
            "Only applies when dumping a database to YAML."
        )
    )
    parser.add_argument(
        "--media-root", help="Filesystem path to directory where images and media are saved"
    )
    parser.add_argument("--media-url", help="URL for media root; must end in slash if non-empty")
    parser.add_argument("--plugin", '-p', action='append', help="Plugin module")
    parser.add_argument(
        "--output", '-o', default='./',
        help="Relative path to output directory",
    )
    parser.add_argument(
        "--verbosity", '-v', default='INFO',
        help="Level to set the root logging module to",
    )
    parser.add_argument(
        "--logging-format", default='%(asctime)s - %(levelname)s: %(message)s',
        help="Default format used when logging",
    )
    parser.add_argument(
        "--database-config", default='{}',
        help="A JSON string in the format {database_id: {sorts: {...}, filter: {...}}}. "
             "Will get passed to the Notion API directly"
    )

    # TODO: Add the ability to dump out a "schema" file that contains the schema
    # for a set of databases

    # TODO: Add the ability to export everything as a sqlite file

    args = parser.parse_args(raw_args)

    logging_level = logging.__dict__[args.verbosity]
    logging.basicConfig(format=args.logging_format, level=logging_level)
    global logger
    logger = logging.getLogger(__name__)

    if access_token is None:
        logger.critical('No NOTION_ACCESS_TOKEN environment variable is set')
        return 1

    object_id = notion.id_from_share_link(args.object_id)
    media_root = args.media_root or args.output

    database_config = database_config_json_to_dict(args.database_config)
    valid_database_config = database_config is not None
    if not valid_database_config:
        logger.critical(
            'Database config validation failed. Please make sure you pass in '
            'a JSON string with the format {database_id: {sorts: {...}, filter: {...}}}'
        )
        return 1

    client = notion.Client(
        access_token,
        media_root,
        args.media_url,
        plugins=args.plugin,
        content_property=args.content_property,
        id_property=args.id_property,
        database_config=database_config,
    )

    node = client.get_page_or_database(object_id)

    # TODO: in the future, determining the natural keys for each row in the
    # database and calculate them up-front; prune out any pages where the
    # natural key is empty. Furthermore, add duplicate handling here.

    if isinstance(node, Database) and args.format == 'markdown':
        export_database_as_markdown_files(node, options=args)
    elif isinstance(node, Database) and args.format == 'yaml':
        print(node.to_yaml())
    elif isinstance(node, Database) and args.format == 'yaml-related':
        export_related_databases(node, options=args)
    elif isinstance(node, Page):
        print(node.to_markdown())
    elif node is None:
        msg = (
            "Unable to find database or page with id %s. "
            "Perhaps its not shared with the integration?"
        )
        logger.error(msg, object_id)
        return 2

    return 0


def database_config_json_to_dict(config_json):
    try:
        config = json.loads(config_json)
    except json.JSONDecodeError:
        return None
    if not validate_database_config(config):
        return None
    return config


def validate_database_config(config):
    try:
        for database_id, config_values in config.items():
            database_id_correct_length = len(database_id) == 32
            if not database_id_correct_length:
                return False
            for key, values in config_values.items():
                if key not in ["sorts", "filter"]:
                    return False
                if not isinstance(values, dict) and not isinstance(values, list):
                    return False
    except AttributeError:
        return False
    return True


def export_database_as_markdown_files(database, options):
    os.makedirs(options.output, exist_ok=True)
    seen_file_names = set()
    counts = {'unnamed': 0, 'duplicate': 0}
    for page in database.children:
        if page.filename:
            if page.filename not in seen_file_names:
                seen_file_names.add(page.filename)
                with open(os.path.join(options.output, f"{page.filename}.md"), 'w') as f:
                    f.write(page.to_markdown())
            else:
                logger.warning('Skipping page named "%s" since it has been used', page.filename)
                counts['duplicate'] += 1
        else:
            counts['unnamed'] += 1
    for key, count in counts.items():
        if count > 0:
            logger.info("%d %s page(s) skipped", count, key)


def export_related_databases(seed_database, options):
    os.makedirs(options.output, exist_ok=True)

    seen_database_ids = set()
    seen_file_names = set()

    def _export_related_databases(database):
        seen_database_ids.add(database.notion_id)
        if database.filename not in seen_file_names:
            seen_file_names.add(database.filename)
            with open(os.path.join(options.output, f"{database.filename}.yml"), 'w') as f:
                f.write(database.to_yaml())
        else:
            logger.warning('Database name "%s" has been used', database.filename)
        for database_id in database.related_database_ids:
            if database_id not in seen_database_ids:
                try:
                    related_database = database.client.get_database(database_id)
                    _export_related_databases(related_database)
                except APIResponseError as err:
                    if err.code == APIErrorCode.ObjectNotFound:
                        msg = 'Skipping database with id "%s" due to lack of permissions'
                        logger.warning(msg, database_id)
                    else:
                        raise err

    _export_related_databases(seed_database)


if __name__ == "__main__":
    cli_main()



from pandoc.types import Str

from n2y.utils import process_notion_date, processed_date_to_plain_text


class Mention:
    def __init__(self, client, notion_data, plain_text):
        self.client = client
        self.plain_text = plain_text
        self.notion_type = notion_data["type"]


class UserMention(Mention):
    def __init__(self, client, notion_data, plain_text):
        super().__init__(client, notion_data, plain_text)
        self.user = client.wrap_notion_user(notion_data["user"])

    def to_pandoc(self):
        return [Str(self.user.name)]


class PageMention(Mention):
    def __init__(self, client, notion_data, plain_text):
        super().__init__(client, notion_data, plain_text)
        self.notion_page_id = notion_data["page"]["id"]

    def to_pandoc(self):
        # TODO: if the page is being exported too, then make this a relative
        # URL to that page
        return [Str(self.plain_text)]


class DatabaseMention(Mention):
    def __init__(self, client, notion_data, plain_text):
        super().__init__(client, notion_data, plain_text)
        self.notion_database_id = notion_data["database"]["id"]

    def to_pandoc(self):
        return [Str(self.plain_text)]


class DateMention(Mention):
    def __init__(self, client, notion_data, plain_text):
        super().__init__(client, notion_data, plain_text)
        self.processed_date = process_notion_date(notion_data["date"])

    def to_pandoc(self):
        # TODO: Consider just using plain_text here
        plain_text_date = processed_date_to_plain_text(self.processed_date)
        return [Str(plain_text_date)]


class LinkPreviewMention(Mention):
    def __init__(self, client, notion_data, plain_text):
        super().__init__(client, notion_data, plain_text)
        self.url = notion_data["link_preview"]["url"]

    def to_pandoc(self):
        return [Str(self.url)]


# TODO: handle template button date and user mentions

DEFAULT_MENTIONS = {
    "user": UserMention,
    "page": PageMention,
    "database": DatabaseMention,
    "date": DateMention,
    "link_preview": LinkPreviewMention,
}



from datetime import datetime
import logging

from n2y.utils import fromisoformat, process_notion_date, processed_date_to_plain_text


logger = logging.getLogger(__name__)


class PropertyValue:
    def __init__(self, client, notion_data):
        self.client = client
        self.notion_property_id = notion_data['id']
        self.notion_type = notion_data['type']

    def to_value(self):
        raise NotImplementedError()


class TitlePropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        # TODO: handle the case when there are more than 25 rich text items in the property
        # See https://developers.notion.com/reference/retrieve-a-page-property
        super().__init__(client, notion_data)
        self.rich_text = client.wrap_notion_rich_text_array(notion_data['title'])

    def to_value(self):
        # Notion allows styling of the title, however, in their UI they display
        # the title property without any styling. Thus, if you copy/paste styled
        # text into a title this styling can be hidden and can re-appear after
        # the document conversion. To avoid this surprise, we only generate
        # plain text here.
        return self.rich_text.to_plain_text()


class TextPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        # TODO: handle the case when there are more than 25 rich text items in the property
        # See https://developers.notion.com/reference/retrieve-a-page-property
        super().__init__(client, notion_data)
        self.rich_text = client.wrap_notion_rich_text_array(notion_data['rich_text'])

    def to_value(self):
        return self.rich_text.to_markdown()


class NumberPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.number = notion_data['number']

    def to_value(self):
        return self.number


class SelectPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        notion_select = notion_data['select']
        if notion_select is not None:
            self.notion_option_id = notion_select['id']
            self.name = notion_select['name']
            self.color = notion_select['color']
        else:
            self.notion_option_id = None
            self.name = None
            self.color = None

    def to_value(self):
        # Note: the Notion UI shouldn't allow you to have two options with the
        # same name
        return self.name


class MultiSelectPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.options = [MultiSelectOption(self.client, no) for no in notion_data['multi_select']]

    def to_value(self):
        # Note: the Notion UI shouldn't allow you to have two options with the
        # same name
        return [o.name for o in self.options]


class MultiSelectOption:
    def __init__(self, client, notion_option):
        self.client = client
        self.notion_id = notion_option['id']
        self.name = notion_option['name']
        self.color = notion_option['color']


class DatePropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        # TODO: handle timezones
        super().__init__(client, notion_data)
        self.value = process_notion_date(notion_data['date'])

    def to_value(self):
        return self.value

    def to_plain_text(self):
        return processed_date_to_plain_text(self.value)


class PeoplePropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.people = [client.wrap_notion_user(nu) for nu in notion_data['people']]

    def to_value(self):
        return [u.to_value() for u in self.people]


class FilesPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.files = [client.wrap_notion_file(nf) for nf in notion_data['files']]

    def to_value(self):
        return [f.to_value() for f in self.files]


class CheckboxPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.checkbox = notion_data['checkbox']

    def to_value(self):
        return self.checkbox


class UrlPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.url = notion_data['url']

    def to_value(self):
        return self.url


class EmailPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.email = notion_data['email']

    def to_value(self):
        return self.email


class PhoneNumberPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.phone_number = notion_data['phone_number']

    def to_value(self):
        return self.phone_number


class FormulaPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        # TODO: set other attributes
        notion_formula = notion_data["formula"]
        if notion_formula["type"] == "date":
            self.value = process_notion_date(notion_formula["date"])
        else:
            self.value = notion_formula[notion_formula["type"]]

    def to_value(self):
        return self.value


class RelationPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        # TODO: handle the case when there are more than 25 pages in the relation
        # See https://developers.notion.com/reference/retrieve-a-page-property
        super().__init__(client, notion_data)
        self.ids = [related["id"] for related in notion_data["relation"]]

    def to_value(self):
        return self.ids


class RollupPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        # TODO: handle the case when the rollup needs to be paginated
        # See https://developers.notion.com/reference/retrieve-a-page-property
        super().__init__(client, notion_data)
        notion_rollup = notion_data["rollup"]
        self.function = notion_rollup["function"]
        if notion_rollup["type"] == "date":
            self.value = process_notion_date(notion_rollup['date'])
        else:
            self.value = notion_rollup[notion_rollup["type"]]
        # TODO: handle arrays of dates

    def to_value(self):
        return self.value


class CreatedTimePropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.created_time = fromisoformat(notion_data['created_time'])

    def to_value(self):
        return datetime.isoformat(self.created_time)


class CreatedByPropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.created_by = client.wrap_notion_user(notion_data['created_by'])

    def to_value(self):
        return self.created_by.to_value()


class LastEditedTimePropertyValue(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.last_edited_time = fromisoformat(notion_data['last_edited_time'])

    def to_value(self):
        return datetime.isoformat(self.last_edited_time)


class LastEditedBy(PropertyValue):
    def __init__(self, client, notion_data):
        super().__init__(client, notion_data)
        self.last_edited_by = client.wrap_notion_user(notion_data['last_edited_by'])

    def to_value(self):
        return self.last_edited_by.to_value()


DEFAULT_PROPERTY_VALUES = {
    'title': TitlePropertyValue,
    'rich_text': TextPropertyValue,
    'number': NumberPropertyValue,
    'select': SelectPropertyValue,
    'multi_select': MultiSelectPropertyValue,
    'date': DatePropertyValue,
    'people': PeoplePropertyValue,
    'files': FilesPropertyValue,
    'checkbox': CheckboxPropertyValue,
    'url': UrlPropertyValue,
    'email': EmailPropertyValue,
    'phone_number': PhoneNumberPropertyValue,
    'formula': FormulaPropertyValue,
    'relation': RelationPropertyValue,
    'rollup': RollupPropertyValue,
    'created_time': CreatedTimePropertyValue,
    'created_by': CreatedByPropertyValue,
    'last_edited_time': LastEditedTimePropertyValue,
    'last_edited_by': LastEditedBy,
}
